﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.IO;
using System.Reflection;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;
using uDrawLib;

namespace uDrawTablet
{
  public partial class Preferences : Form
  {
    #region P/Invoke Crud

    [DllImport("kernel32")]
    private static extern long WritePrivateProfileString(string lPAppName, string lpKeyName, string lpString, string lpFileName);

    [DllImport("kernel32")]
    private static extern int GetPrivateProfileString(string lpAppName, string lpKeyName, string lpDefault, StringBuilder lpReturnedString, int nSize, string lpFilePath);

    #endregion

    #region Declarations

    private const string _INI_FILE = "Settings.ini";
    private const string _DEFAULT_SECTION = "Default";
    private const string _KEY_PEN_THRESHOLD = "PenThreshold";
    private NotifyIcon _icon;
    private ContextMenu _menu;
    private bool _inPreferences;

    #endregion

    #region Constructors / Teardown

    public Preferences()
    {
      InitializeComponent();

      _menu = new ContextMenu();
      _menu.MenuItems.Add("Preferences", OnPreferences);
      _menu.MenuItems.Add("Exit", OnExit);

      _icon = new NotifyIcon();
      _icon.Text = Assembly.GetExecutingAssembly().GetName().Name;
      _icon.Icon = Icon.ExtractAssociatedIcon(Assembly.GetExecutingAssembly().Location);
      _icon.ContextMenu = _menu;
      _icon.Visible = true;

      if (!uDrawTabletDevice.IsDetected())
      {
        MessageBox.Show(this, "PS3 uDraw tablet was not detected!", "Not Detected", MessageBoxButtons.OK,
          MessageBoxIcon.Exclamation);

        Application.Exit();
      }
      else
      {
        MouseInterface.Start(this);
      }
    }

    #endregion

    #region Public Methods

    public void ShowPreferences()
    {
      OnPreferences(this, EventArgs.Empty);
    }

    public void LoadPreferences()
    {
      var sb = new StringBuilder(255);
      GetPrivateProfileString(_DEFAULT_SECTION, _KEY_PEN_THRESHOLD, MouseInterface.PenPressureThreshold.ToString(), sb, sb.Capacity,
        Path.Combine(Directory.GetCurrentDirectory(), _INI_FILE));
      int threshold = MouseInterface.PenPressureThreshold; int.TryParse(sb.ToString(), out threshold);
      MouseInterface.PenPressureThreshold = threshold;
    }

    #endregion

    #region Overrides

    protected override void SetVisibleCore(bool value)
    {
      if (!this.IsHandleCreated)
      {
        value = false;
        CreateHandle();
      }

      base.SetVisibleCore(value);
    }

    protected override void Dispose(bool isDisposing)
    {
      if (isDisposing)
      {
        if (components != null)
          components.Dispose();

        //Get rid of the icon
        _icon.Dispose();
      }

      base.Dispose(isDisposing);
    }

    #endregion

    #region Event Handlers

    private void btnSave_Click(object sender, EventArgs e)
    {
      //Save preferences
      MouseInterface.PenPressureThreshold = trbPenClick.Maximum - trbPenClick.Value + trbPenClick.Minimum;
      WritePrivateProfileString(_DEFAULT_SECTION, _KEY_PEN_THRESHOLD, MouseInterface.PenPressureThreshold.ToString(),
        Path.Combine(Directory.GetCurrentDirectory(), _INI_FILE));

      this.Close();
    }

    private void btnCancel_Click(object sender, EventArgs e)
    {
      this.Close();
    }

    private void Main_FormClosing(object sender, FormClosingEventArgs e)
    {
      if (_inPreferences)
      {
        _inPreferences = false;

        e.Cancel = true;
        this.Hide();
      }
    }

    #endregion

    #region Local Methods

    private void _ShowPreferences()
    {
      const int TICK_FREQUENCY = 0x10;
      const int PEN_PRESSURE_MINIMUM = 0x70;
      const int PEN_PRESSURE_MAXIMUM = 0xFF;

      if (this.InvokeRequired)
      {
        this.Invoke(new MethodInvoker(_ShowPreferences));
      }
      else
      {
        //Load preferences
        LoadPreferences();
        trbPenClick.Minimum = PEN_PRESSURE_MINIMUM;
        trbPenClick.Maximum = PEN_PRESSURE_MAXIMUM;
        trbPenClick.TickFrequency = -TICK_FREQUENCY;
        trbPenClick.Value = trbPenClick.Maximum - MouseInterface.PenPressureThreshold + trbPenClick.Minimum;

        this.Show();
      }
    }

    private void OnPreferences(object sender, EventArgs e)
    {
      _inPreferences = true;

      _ShowPreferences();
    }

    private void OnExit(object sender, EventArgs e)
    {
      try
      {
        _inPreferences = false;
        MouseInterface.Stop();
      }
      finally
      {
        Application.Exit();
      }
    }

    #endregion
  }
}
